from pygame_gui.elements.text.html_parser import TextHTMLParser, CharStyle, TextLineContext
from pygame_gui.elements.text.text_effects import TextBoxEffect, TypingAppearEffect, FadeOutEffect, FadeInEffect
